package ru.amse.cat.evlarchick.textstructure;

import java.util.*;
import java.io.Reader;
import java.io.IOException;

public class TextStructure implements ITextStructure {
    private String myOriginalText;
    private List<ISegmentWithTranslation> mySegments;
    private List<ISegment> myDeadZones;

    public TextStructure(String originalText) {
        this(originalText, Collections.<ISegmentWithTranslation>emptyList());
    }

    public TextStructure(String originalText, String filepath) {
        this(originalText, Collections.<ISegmentWithTranslation>emptyList(), filepath);
    }

    // usable constructor?????
    public TextStructure(String originalText, List<ISegmentWithTranslation> list) {
        this(originalText, list, null);
    }

    public TextStructure(String originalText, List<ISegmentWithTranslation> list, String filePath) {
        myOriginalText = originalText;
        mySegments = new ArrayList<ISegmentWithTranslation>(list);
        myDeadZones = new ArrayList<ISegment>();
    }

    // simple realisation
    public ISegment getSegmentByIndex(int index) {
        for (ISegmentWithTranslation s : mySegments) {
            if (s.getStart() <= index && s.getEnd() > index) {
                return s;
            }
        }
        for (ISegment s : myDeadZones) {
            if (s.getStart() <= index && s.getEnd() > index) {
                return s;
            }
        }
        return null;
    }

    public String getOriginalText() {
        return myOriginalText;
    }

    public List<ISegmentWithTranslation> getSegments() {
        return Collections.unmodifiableList(mySegments);
    }

    public Reader getTranslationReader() {
        return new TranslationReader();
    }

    // simple realisation
    public void addDeadZone(ISegment segment) {
        for (int i = 0; i < myDeadZones.size(); i++) {
            if (myDeadZones.get(i).getStart() >= segment.getEnd()) {
                myDeadZones.add(i, segment);
                return;
            }
        }
        myDeadZones.add(segment);
    }

    public List<ISegment> getCommonSegmentList() {
        Iterator<ISegment> it1 = myDeadZones.iterator();
        Iterator<ISegmentWithTranslation> it2 = mySegments.iterator();

        List<ISegment> commonList = new LinkedList<ISegment>();

        ISegment deadZone = (it1.hasNext()) ? it1.next() : null;
        ISegment segment = (it2.hasNext()) ? it2.next() : null;

        if (!((deadZone == null) && (segment == null))) {
            do {
                if (segment == null)  {
                    commonList.add(deadZone);
                    deadZone = (it1.hasNext()) ? it1.next() : null;
                    if (deadZone == null) {
                        break;
                    }
                } else if(deadZone == null) {
                    commonList.add(segment);
                    segment = (it2.hasNext()) ? it2.next() : null;
                    if (segment == null) {
                        break;
                    }
                } else {
                    if (deadZone.getStart() < segment.getStart()) {
                        commonList.add(deadZone);
                        deadZone = (it1.hasNext()) ? it1.next() : null;
                    } else {
                        commonList.add(segment);
                        segment = (it2.hasNext()) ? it2.next() : null;
                    }
                }
            } while (true);
        }
        return commonList;
    }

    public List<ISegment> getDeadZones() {
        return Collections.unmodifiableList(myDeadZones);
    }

    // simple realisation todo add orginalText bounds
    public boolean isIndicesUseable(ISegment segment, int begin, int end) {
        for (ISegment seg: myDeadZones) {
            if (seg.equals(segment)) {
                continue;
            }

            int dBegin = seg.getStart();
            int dEnd = seg.getEnd();

            if ((begin >= end) || (begin >= dBegin && begin < dEnd) ||
                    (end > dBegin && end <= dEnd) ||
                    (begin < dBegin && end >= dEnd)) {
                return false;

            }
        }

        for (ISegment seg: mySegments) {
            if (seg.equals(segment)) {
                continue;
            }

            int dBegin = seg.getStart();
            int dEnd = seg.getEnd();


            if ((begin >= end) || (begin >= dBegin && begin < dEnd) ||
                    (end > dBegin && end <= dEnd) ||
                    (begin < dBegin && end >= dEnd)) {
                return false;
            }
        }

        return true;
    }

    // simple realization
    public void addSegmentWithTranslation(ISegmentWithTranslation segmentWithTranslation) {
        for (int i = 0; i < mySegments.size(); i++) {
            if (mySegments.get(i).getStart() >= segmentWithTranslation.getEnd()) {
                mySegments.add(i, segmentWithTranslation);
                return;
            }
        }
        mySegments.add(segmentWithTranslation);
    }

    public void setText(String s) {
        myOriginalText = s;
    }

    private class TranslationReader extends Reader {
        private int myLength;
        private int myReadAlredy;
        private int myCurrentRead;
        private String myCurrentString;
        private boolean myInSegment;
        private int myCounter;
        private Iterator<ISegmentWithTranslation> myIterator;
        private ISegmentWithTranslation myCurrentSegment;


        TranslationReader() {
            init();
        }

        private void init() {
            myReadAlredy = 0;
            myCurrentRead = 0;
            myInSegment = true;
            myCounter = 0;
            myCurrentSegment = null;

            myLength = isAvailable();
            if (myLength > 0) {
                myIterator = mySegments.iterator();
                myCurrentString = nextString();
            }
        }

        private int isAvailable() {
            int length = 0;
            int currentEnd = 0;

            for (ISegmentWithTranslation segment: mySegments) {
                length += segment.getStart() - currentEnd;
                length += segment.getTranslation().length();
                currentEnd = segment.getEnd();
            }

            length += myOriginalText.length() - currentEnd;
            return length;
        }

        public int read(char cbuf[], int off, int len) {
            if ((off < 0) || (off > cbuf.length) || (len < 0) ||
                    ((off + len) > cbuf.length) || ((off + len) < 0)) {
                throw new IndexOutOfBoundsException();
            } else if (len == 0) {
                return 0;
            }

            if (myReadAlredy >= myLength) {
                return -1;
            }

            int n = Math.min(myLength - myReadAlredy, len);
            int result = n;
            int toRead;

            while ((n - (toRead = (myCurrentString.length() - myCurrentRead))) > 0) {
                myCurrentString.getChars(myCurrentRead, myCurrentRead + toRead, cbuf, off);

                n -= toRead;
                off += toRead;
                myReadAlredy += toRead;

                myCurrentString = nextString();
                myCurrentRead = 0;
            }

            myCurrentString.getChars(myCurrentRead, n + myCurrentRead, cbuf, off);
            myCurrentRead += n;
            myReadAlredy += n;

            return result;
        }

        private String nextString() {
            String result;
            if (myInSegment) {
                myCurrentSegment = (myIterator.hasNext()) ? myIterator.next() : null;

                if (myCurrentSegment == null) {
                    myInSegment = false;
                    result = myOriginalText.substring(myCounter, myOriginalText.length());
                    myCounter += result.length();
                } else {
                    if (myCurrentSegment.getStart() == myCounter) {
                        result = myCurrentSegment.getTranslation();
                        myCounter += myCurrentSegment.getEnd() - myCurrentSegment.getStart();
                    } else {
                        myInSegment = false;
                        result = myOriginalText.substring(myCounter,
                                myCurrentSegment.getStart());
                        myCounter = myCurrentSegment.getStart();
                    }
                }

            } else {
                myInSegment = true;
                result = myCurrentSegment.getTranslation();
                myCounter = myCurrentSegment.getEnd();
            }
            return result;
        }

        public void close() throws IOException {
            myLength = 0;
        }

        public void reset() {
            init();
        }
    }
}
